from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest
from packaging.version import Version

from scikit_build_core.cmake import CMake, CMaker

DIR = Path(__file__).parent.absolute()


@pytest.fixture(scope="session")
def config(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("build")

    build_dir = tmp_path / "build"

    cmake = CMake.default_search(minimum_version=Version("3.15"))
    config = CMaker(
        cmake,
        source_dir=DIR / "packages/simple_pure",
        build_dir=build_dir,
        build_type="Release",
    )

    config.configure()

    config.build()

    return config


@pytest.mark.compile
@pytest.mark.configure
@pytest.mark.skipif(
    sys.platform.startswith("win32") or sys.platform.startswith("cygwin"),
    reason="Paths different in build dir on Windows",
)
def test_bin_in_config(config):
    result = subprocess.run(
        [str(config.build_dir / "simple_pure")],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0
    assert result.stdout == "0 one 2 three \n"


# TODO: figure out why gmake is reporting no rule to make simple_pure.cpp
@pytest.mark.compile
@pytest.mark.configure
@pytest.mark.xfail(
    sys.platform.startswith("cygwin"), strict=False, reason="No idea why this fails"
)
def test_install(config):
    install_dir = config.build_dir.parent / "install"
    config.install(install_dir)

    result = subprocess.run(
        [str(install_dir / "bin/simple_pure")],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0
    assert result.stdout == "0 one 2 three \n"


@pytest.mark.configure
def test_variable_defined(tmp_path, capfd):
    build_dir = tmp_path / "build"

    cmake = CMake.default_search(minimum_version=Version("3.15"))
    config = CMaker(
        cmake,
        source_dir=DIR / "packages/simple_pure",
        build_dir=build_dir,
        build_type="Release",
    )
    config.init_cache({"SKBUILD": True})
    config.configure(defines={"SKBUILD2": True})

    out = capfd.readouterr().out
    assert "SKBUILD is defined to ON" in out
    assert "SKBUILD2 is defined to ON" in out



from __future__ import annotations

import textwrap

import pytest

import scikit_build_core.settings.skbuild_read_settings
from scikit_build_core.settings.skbuild_read_settings import SettingsReader


def test_skbuild_settings_default(tmp_path):
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text("", encoding="utf-8")

    config_settings: dict[str, list[str] | str] = {}

    settings_reader = SettingsReader(pyproject_toml, config_settings)
    settings = settings_reader.settings
    assert list(settings_reader.unrecognized_options()) == []

    assert settings.ninja.minimum_version == "1.5"
    assert settings.ninja.make_fallback
    assert settings.cmake.minimum_version == "3.15"
    assert settings.cmake.args == []
    assert settings.cmake.define == {}
    assert not settings.cmake.verbose
    assert settings.cmake.build_type == "Release"
    assert settings.logging.level == "WARNING"
    assert settings.sdist.include == []
    assert settings.sdist.exclude == []
    assert settings.sdist.reproducible
    assert settings.wheel.packages is None
    assert settings.wheel.py_api == ""
    assert not settings.wheel.expand_macos_universal_tags
    assert settings.backport.find_python == "3.24"
    assert settings.strict_config
    assert not settings.experimental
    assert settings.minimum_version is None


def test_skbuild_settings_envvar(tmp_path, monkeypatch):
    monkeypatch.setattr(
        scikit_build_core.settings.skbuild_read_settings, "__version__", "0.1.0"
    )

    monkeypatch.setenv("SKBUILD_NINJA_MINIMUM_VERSION", "1.1")
    monkeypatch.setenv("SKBUILD_NINJA_MAKE_FALLBACK", "0")
    monkeypatch.setenv("SKBUILD_CMAKE_MINIMUM_VERSION", "3.16")
    monkeypatch.setenv("SKBUILD_CMAKE_ARGS", "-DFOO=BAR;-DBAR=FOO")
    monkeypatch.setenv("SKBUILD_CMAKE_DEFINE", "a=1;b=2")
    monkeypatch.setenv("SKBUILD_CMAKE_BUILD_TYPE", "Debug")
    monkeypatch.setenv("SKBUILD_LOGGING_LEVEL", "DEBUG")
    monkeypatch.setenv("SKBUILD_SDIST_INCLUDE", "a;b; c")
    monkeypatch.setenv("SKBUILD_SDIST_EXCLUDE", "d;e;f")
    monkeypatch.setenv("SKBUILD_SDIST_REPRODUCIBLE", "OFF")
    monkeypatch.setenv("SKBUILD_WHEEL_PACKAGES", "j; k; l")
    monkeypatch.setenv("SKBUILD_WHEEL_PY_API", "cp39")
    monkeypatch.setenv("SKBUILD_WHEEL_EXPAND_MACOS_UNIVERSAL_TAGS", "True")
    monkeypatch.setenv("SKBUILD_BACKPORT_FIND_PYTHON", "0")
    monkeypatch.setenv("SKBUILD_STRICT_CONFIG", "0")
    monkeypatch.setenv("SKBUILD_EXPERIMENTAL", "1")
    monkeypatch.setenv("SKBUILD_MINIMUM_VERSION", "0.1")
    monkeypatch.setenv("SKBUILD_CMAKE_VERBOSE", "TRUE")

    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text("", encoding="utf-8")

    config_settings: dict[str, list[str] | str] = {}

    settings_reader = SettingsReader(pyproject_toml, config_settings)
    settings = settings_reader.settings
    assert list(settings_reader.unrecognized_options()) == []

    assert settings.ninja.minimum_version == "1.1"
    assert settings.cmake.minimum_version == "3.16"
    assert settings.cmake.args == ["-DFOO=BAR", "-DBAR=FOO"]
    assert settings.cmake.define == {"a": "1", "b": "2"}
    assert settings.cmake.verbose
    assert settings.cmake.build_type == "Debug"
    assert not settings.ninja.make_fallback
    assert settings.logging.level == "DEBUG"
    assert settings.sdist.include == ["a", "b", "c"]
    assert settings.sdist.exclude == ["d", "e", "f"]
    assert not settings.sdist.reproducible
    assert settings.wheel.packages == ["j", "k", "l"]
    assert settings.wheel.py_api == "cp39"
    assert settings.wheel.expand_macos_universal_tags
    assert settings.backport.find_python == "0"
    assert not settings.strict_config
    assert settings.experimental
    assert settings.minimum_version == "0.1"


def test_skbuild_settings_config_settings(tmp_path, monkeypatch):
    monkeypatch.setattr(
        scikit_build_core.settings.skbuild_read_settings, "__version__", "0.1.0"
    )

    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text("", encoding="utf-8")

    config_settings: dict[str, str | list[str]] = {
        "ninja.minimum-version": "1.2",
        "ninja.make-fallback": "False",
        "cmake.minimum-version": "3.17",
        "cmake.args": ["-DFOO=BAR", "-DBAR=FOO"],
        "cmake.define.a": "1",
        "cmake.define.b": "2",
        "cmake.verbose": "true",
        "cmake.build-type": "Debug",
        "logging.level": "INFO",
        "sdist.include": ["a", "b", "c"],
        "sdist.exclude": "d;e;f",
        "sdist.reproducible": "false",
        "wheel.packages": ["j", "k", "l"],
        "wheel.py-api": "cp39",
        "wheel.expand-macos-universal-tags": "True",
        "backport.find-python": "",
        "strict-config": "false",
        "experimental": "1",
        "minimum-version": "0.1",
    }

    settings_reader = SettingsReader(pyproject_toml, config_settings)
    settings = settings_reader.settings
    assert list(settings_reader.unrecognized_options()) == []

    assert settings.ninja.minimum_version == "1.2"
    assert not settings.ninja.make_fallback
    assert settings.cmake.minimum_version == "3.17"
    assert settings.cmake.args == ["-DFOO=BAR", "-DBAR=FOO"]
    assert settings.cmake.define == {"a": "1", "b": "2"}
    assert settings.cmake.verbose
    assert settings.cmake.build_type == "Debug"
    assert settings.logging.level == "INFO"
    assert settings.sdist.include == ["a", "b", "c"]
    assert settings.sdist.exclude == ["d", "e", "f"]
    assert not settings.sdist.reproducible
    assert settings.wheel.packages == ["j", "k", "l"]
    assert settings.wheel.py_api == "cp39"
    assert settings.wheel.expand_macos_universal_tags
    assert settings.backport.find_python == ""
    assert not settings.strict_config
    assert settings.experimental
    assert settings.minimum_version == "0.1"


def test_skbuild_settings_pyproject_toml(tmp_path, monkeypatch):
    monkeypatch.setattr(
        scikit_build_core.settings.skbuild_read_settings, "__version__", "0.1.0"
    )
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """\
            [tool.scikit-build]
            ninja.minimum-version = "1.3"
            ninja.make-fallback = false
            cmake.minimum-version = "3.18"
            cmake.args = ["-DFOO=BAR", "-DBAR=FOO"]
            cmake.define = {a = "1", b = "2"}
            cmake.build-type = "Debug"
            cmake.verbose = true
            logging.level = "ERROR"
            sdist.include = ["a", "b", "c"]
            sdist.exclude = ["d", "e", "f"]
            sdist.reproducible = false
            wheel.packages = ["j", "k", "l"]
            wheel.py-api = "cp39"
            wheel.expand-macos-universal-tags = true
            backport.find-python = "3.18"
            strict-config = false
            experimental = true
            minimum-version = "0.1"
            """
        ),
        encoding="utf-8",
    )

    config_settings: dict[str, list[str] | str] = {}

    settings_reader = SettingsReader(pyproject_toml, config_settings)
    settings = settings_reader.settings
    assert list(settings_reader.unrecognized_options()) == []

    assert settings.ninja.minimum_version == "1.3"
    assert not settings.ninja.make_fallback
    assert settings.cmake.minimum_version == "3.18"
    assert settings.cmake.args == ["-DFOO=BAR", "-DBAR=FOO"]
    assert settings.cmake.define == {"a": "1", "b": "2"}
    assert settings.cmake.verbose
    assert settings.cmake.build_type == "Debug"
    assert settings.logging.level == "ERROR"
    assert settings.sdist.include == ["a", "b", "c"]
    assert settings.sdist.exclude == ["d", "e", "f"]
    assert not settings.sdist.reproducible
    assert settings.wheel.packages == ["j", "k", "l"]
    assert settings.wheel.py_api == "cp39"
    assert settings.wheel.expand_macos_universal_tags
    assert settings.backport.find_python == "3.18"
    assert not settings.strict_config
    assert settings.experimental
    assert settings.minimum_version == "0.1"


def test_skbuild_settings_pyproject_toml_broken(tmp_path, monkeypatch):
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        textwrap.dedent(
            """\
            [tool.scikit-build]
            cmake.minimum-verison = "3.18"
            ninja.minimum-version = "1.3"
            ninja.make-fallback = false
            logger.level = "ERROR"
            """
        ),
        encoding="utf-8",
    )

    config_settings: dict[str, list[str] | str] = {}

    settings_reader = SettingsReader(pyproject_toml, config_settings)
    assert list(settings_reader.unrecognized_options()) == [
        "tool.scikit-build.cmake.minimum-verison",
        "tool.scikit-build.logger",
    ]

    with pytest.raises(SystemExit):
        settings_reader.validate_may_exit()


def test_skbuild_settings_pyproject_conf_broken(tmp_path):
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text("", encoding="utf-8")

    config_settings: dict[str, str | list[str]] = {
        "cmake.minimum-verison": "3.17",
        "ninja.minimum-version": "1.2",
        "ninja.make-fallback": "False",
        "logger.level": "INFO",
    }

    settings_reader = SettingsReader(pyproject_toml, config_settings)
    assert list(settings_reader.unrecognized_options()) == [
        "cmake.minimum-verison",
        "logger",
    ]

    with pytest.raises(SystemExit):
        settings_reader.validate_may_exit()



from __future__ import annotations

import functools
import os
import shutil
from collections.abc import Generator
from pathlib import Path

import pytest
from packaging.version import Version

from scikit_build_core.cmake import CMake, CMaker
from scikit_build_core.errors import CMakeNotFoundError

DIR = Path(__file__).parent.resolve()


@functools.lru_cache(1)
def cmake_path() -> str:
    try:
        import cmake
    except ModuleNotFoundError:
        cmake_str = shutil.which("cmake")
        print(cmake_str)
        if cmake_str is None:
            pytest.skip("cmake must be installed for this test")
        return os.fspath(Path(cmake_str).resolve())

    return os.fspath(Path(cmake.CMAKE_BIN_DIR) / "cmake")


def configure_args(config: CMaker, *, init: bool = False) -> Generator[str, None, None]:

    yield cmake_path()
    yield f"-S{config.source_dir}"
    yield f"-B{config.build_dir}"

    if init:
        cmake_init = config.build_dir / "CMakeInit.txt"
        yield f"-C{cmake_init}"

    if config.single_config:
        yield f"-DCMAKE_BUILD_TYPE={config.build_type}"

    if config.module_dirs:
        yield "-DCMAKE_MODULE_PATH:PATH={}".format(
            ";".join(str(p).replace("\\", "/") for p in config.module_dirs)
        )

    if config.prefix_dirs:
        yield "-DCMAKE_PREFIX_PATH:PATH={}".format(
            ";".join(str(p).replace("\\", "/") for p in config.prefix_dirs)
        )


@pytest.mark.configure
def test_init_cache(fp, tmp_path):
    fp.register([cmake_path(), "--version"], stdout="3.14.0")

    config = CMaker(
        CMake.default_search(),
        source_dir=DIR / "packages/simple_pure",
        build_dir=tmp_path / "build",
        build_type="Release",
    )
    config.init_cache(
        {"SKBUILD": True, "SKBUILD_VERSION": "1.0.0", "SKBUILD_PATH": config.source_dir}
    )

    cmd = list(configure_args(config, init=True))
    print("Registering:", *cmd)
    fp.register(cmd)
    config.configure()

    cmake_init = config.build_dir / "CMakeInit.txt"
    source_dir_str = str(config.source_dir).replace("\\", "/")
    assert (
        cmake_init.read_text()
        == f"""\
set(SKBUILD ON CACHE BOOL "")
set(SKBUILD_VERSION [===[1.0.0]===] CACHE STRING "")
set(SKBUILD_PATH [===[{source_dir_str}]===] CACHE PATH "")
"""
    )


@pytest.mark.configure
def test_too_old(fp, monkeypatch):
    monkeypatch.setattr(shutil, "which", lambda x: None)
    fp.register([cmake_path(), "--version"], stdout="3.14.0")
    print(cmake_path)

    with pytest.raises(CMakeNotFoundError) as excinfo:
        CMake.default_search(minimum_version=Version("3.15"))
    assert "Could not find CMake with version >= 3.15" in excinfo.value.args[0]


@pytest.mark.configure
def test_cmake_args(tmp_path, fp):
    fp.register([cmake_path(), "--version"], stdout="3.15.0")

    config = CMaker(
        CMake.default_search(),
        source_dir=DIR / "packages/simple_pure",
        build_dir=tmp_path / "build",
        build_type="Release",
    )

    cmd = list(configure_args(config))
    cmd.append("-DSOMETHING=one")
    print("Registering:", *cmd)
    fp.register(cmd)

    config.configure(cmake_args=["-DSOMETHING=one"])

    assert len(fp.calls) == 2


@pytest.mark.configure
def test_cmake_paths(tmp_path, fp):
    fp.register([cmake_path(), "--version"], stdout="3.15.0")

    config = CMaker(
        CMake.default_search(),
        source_dir=DIR / "packages/simple_pure",
        build_dir=tmp_path / "build",
        build_type="Release",
        prefix_dirs=[tmp_path / "prefix"],
        module_dirs=[tmp_path / "module"],
    )

    cmd = list(configure_args(config))
    print("Registering:", *cmd)
    fp.register(cmd)

    config.configure()

    assert len(fp.calls) == 2



from __future__ import annotations

import importlib.util
import os
import shutil
import subprocess
import sys
import types
import warnings
from collections.abc import Generator
from pathlib import Path
from venv import EnvBuilder

if sys.version_info < (3, 8):
    import importlib_metadata as metadata
    from typing_extensions import Literal, overload
else:
    from importlib import metadata
    from typing import overload, Literal

import distlib.wheel
import pytest

DIR = Path(__file__).parent.resolve()
BASE = DIR.parent


@pytest.fixture(scope="session")
def pep518_wheelhouse(tmp_path_factory: pytest.TempPathFactory) -> Path:
    wheelhouse = tmp_path_factory.mktemp("wheelhouse")

    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "wheel",
            "--wheel-dir",
            str(wheelhouse),
            f"{BASE}[pyproject]",
        ],
        check=True,
    )
    packages = [
        "build",
        "pybind11",
        "rich",
        "setuptools",
        "virtualenv",
        "wheel",
    ]

    if importlib.util.find_spec("cmake") is not None:
        packages.append("cmake")

    if importlib.util.find_spec("ninja") is not None:
        packages.append("ninja")

    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "download",
            "-q",
            "-d",
            str(wheelhouse),
            *packages,
        ],
        check=True,
    )
    return wheelhouse


class VEnv(EnvBuilder):
    executable: Path
    env_dir: Path

    def __init__(self, env_dir: Path, *, wheelhouse: Path | None = None) -> None:
        super().__init__(with_pip=True)
        # This warning is mistakenly generated by CPython 3.11.0
        # https://github.com/python/cpython/pull/98743
        with warnings.catch_warnings():
            if sys.version_info[:3] == (3, 11, 0):
                warnings.filterwarnings(
                    "ignore",
                    "check_home argument is deprecated and ignored.",
                    DeprecationWarning,
                )
            self.create(env_dir)
        self.wheelhouse = wheelhouse

    def ensure_directories(
        self, env_dir: str | bytes | os.PathLike[str] | os.PathLike[bytes]
    ) -> types.SimpleNamespace:
        context = super().ensure_directories(env_dir)
        # Store the path to the venv Python interpreter.
        # See https://github.com/mesonbuild/meson-python/blob/8a180be7b4abd7e1939a63d5d59f63197ee27cc7/tests/conftest.py#LL79
        self.executable = Path(context.env_exe)
        self.env_dir = Path(context.env_dir)
        return context

    @overload
    def run(self, *args: str, capture: Literal[True]) -> str:
        ...

    @overload
    def run(self, *args: str, capture: Literal[False] = ...) -> None:
        ...

    def run(self, *args: str, capture: bool = False) -> str | None:
        env = os.environ.copy()
        env["PATH"] = f"{self.executable.parent}{os.pathsep}{env['PATH']}"
        env["VIRTUAL_ENV"] = str(self.env_dir)
        env["PIP_DISABLE_PIP_VERSION_CHECK"] = "ON"
        if self.wheelhouse is not None:
            env["PIP_NO_INDEX"] = "ON"
            env["PIP_FIND_LINKS"] = str(self.wheelhouse)

        if capture:
            return subprocess.run(
                [os.fspath(a) for a in args],
                check=True,
                capture_output=True,
                text=True,
                env=env,
            ).stdout.strip()

        subprocess.run(
            [os.fspath(a) for a in args],
            check=True,
            env=env,
        )
        return None

    def execute(self, command: str) -> str:
        return self.run(str(self.executable), "-c", command, capture=True)

    def module(self, *args: str) -> None:
        return self.run(str(self.executable), "-m", *args)

    def install(self, *args: str) -> None:
        self.module("pip", "install", *args)


@pytest.fixture
def virtualenv(tmp_path: Path, pep518_wheelhouse: Path) -> Generator[VEnv, None, None]:
    path = tmp_path / "venv"
    try:
        yield VEnv(path, wheelhouse=pep518_wheelhouse)
    finally:
        shutil.rmtree(path)


def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    for item in items:
        # Ensure all tests using virtualenv are marked as such
        if "virtualenv" in getattr(item, "fixturenames", ()):
            item.add_marker(pytest.mark.virtualenv)


def pytest_report_header() -> str:
    interesting_packages = {
        "build",
        "distlib",
        "packaging",
        "pathspec",
        "pip",
        "pybind11",
        "pyproject_metadata",
        "rich",
        "setuptools",
        "virtualenv",
        "wheel",
    }
    valid = []
    for package in interesting_packages:
        try:
            version = metadata.version(package)  # type: ignore[no-untyped-call]
        except ModuleNotFoundError:
            continue
        valid.append(f"{package}=={version}")
    reqs = " ".join(sorted(valid))
    pkg_line = f"installed packages of interest: {reqs}"

    wheel = distlib.wheel.Wheel()
    wheel.abi = [distlib.wheel.ABI]
    wheel.pyver = [distlib.wheel.IMPVER]
    wheel.arch = [distlib.wheel.ARCH]
    wheel_line = f"default wheelname: {wheel.filename}"

    return "\n".join([pkg_line, wheel_line])



# pylint: disable=duplicate-code

import builtins
import json
from pathlib import Path
from typing import Any, Callable, Dict, Type, TypeVar

import cattr
import cattr.preconf.json

from .model.cache import Cache
from .model.cmakefiles import CMakeFiles
from .model.codemodel import CodeModel, Target
from .model.index import Index, Reply

T = TypeVar("T")

__all__ = ["make_converter", "load_reply_dir"]


def to_path(path: str, _: Type[Path]) -> Path:
    return Path(path)


def make_converter(base_dir: Path) -> cattr.preconf.json.JsonConverter:
    converter = cattr.preconf.json.make_converter()
    converter.register_structure_hook(Path, to_path)

    st_hook = cattr.gen.make_dict_structure_fn(
        Reply,
        converter,
        codemodel_v2=cattr.gen.override(rename="codemodel-v2"),
        cache_v2=cattr.gen.override(rename="cache-v2"),
        cmakefiles_v1=cattr.gen.override(rename="cmakeFiles-v1"),
        toolchains_v1=cattr.gen.override(rename="toolchains-v1"),
    )
    converter.register_structure_hook(Reply, st_hook)

    def from_json_file(with_path: Dict[str, Any], t: Type[T]) -> T:
        if with_path["jsonFile"] is None:
            return converter.structure_attrs_fromdict({}, t)
        path = base_dir / Path(with_path["jsonFile"])
        raw = json.loads(path.read_text(encoding="utf-8"))
        return converter.structure_attrs_fromdict(raw, t)

    converter.register_structure_hook(CodeModel, from_json_file)
    converter.register_structure_hook(Target, from_json_file)
    converter.register_structure_hook(Cache, from_json_file)
    converter.register_structure_hook(CMakeFiles, from_json_file)
    return converter


def load_reply_dir(reply_dir: Path) -> Index:
    converter = make_converter(reply_dir)
    indexes = sorted(reply_dir.glob("index-*"))
    if not indexes:
        raise IndexError(f"index file not found in {reply_dir}")
    index_file = indexes[-1]
    return converter.loads(index_file.read_text(), Index)


if __name__ == "__main__":
    import argparse

    rich_print: Callable[[object], None]
    try:
        from rich import print as rich_print
    except ModuleNotFoundError:
        rich_print = builtins.print

    parser = argparse.ArgumentParser()
    parser.add_argument("reply_dir", type=Path, help="Path to the reply directory")
    args = parser.parse_args()

    reply = Path(args.reply_dir)
    rich_print(load_reply_dir(reply))



from __future__ import annotations

import shutil
import sys
import tempfile
from collections.abc import Sequence
from pathlib import Path

import distlib.wheel
import packaging.utils
from packaging.version import Version
from pyproject_metadata import StandardMetadata

from .. import __version__
from .._compat import tomllib
from .._logging import logger, rich_print
from ..builder.builder import Builder
from ..builder.wheel_tag import WheelTag
from ..cmake import CMake, CMaker
from ..settings.skbuild_read_settings import SettingsReader
from ._file_processor import each_unignored_file
from ._init import setup_logging

__all__: list[str] = ["build_wheel"]


def __dir__() -> list[str]:
    return __all__


class DistWheel(distlib.wheel.Wheel):  # type: ignore[misc]
    wheel_version = (1, 0)


def _copy_python_packages_to_wheel(
    *,
    packages: Sequence[str] | None,
    name: str,
    platlib_dir: Path,
    include: Sequence[str],
    exclude: Sequence[str],
) -> None:
    if packages is None:
        # Auto package discovery
        packages = []
        for base_path in (Path("src"), Path(".")):
            path = base_path / name
            if path.is_dir() and (
                (path / "__init__.py").is_file() or (path / "__init__.pyi").is_file()
            ):
                logger.info("Discovered Python package at {}", path)
                packages += [str(path)]
                break
        else:
            logger.debug("Didn't find a Python package for {}", name)

    for package in packages:
        source_package = Path(package)
        base_path = source_package.parent
        for filepath in each_unignored_file(
            source_package,
            include=include,
            exclude=exclude,
        ):
            package_dir = platlib_dir / filepath.relative_to(base_path)
            if not package_dir.is_file():
                package_dir.parent.mkdir(exist_ok=True, parents=True)
                shutil.copyfile(filepath, package_dir)


def _write_wheel_metadata(
    *,
    packages_dir: Path,
    metadata: StandardMetadata,
) -> None:
    name = packaging.utils.canonicalize_name(metadata.name).replace("-", "_")
    version = str(metadata.version)
    dist_info = packages_dir / Path(f"{name}-{version}.dist-info")
    dist_info.mkdir(exist_ok=False)
    with dist_info.joinpath("METADATA").open("wb") as f:
        f.write(bytes(metadata.as_rfc822()))
    with dist_info.joinpath("entrypoints.txt").open("w", encoding="utf_8") as f:
        ep = metadata.entrypoints.copy()
        ep["console_scripts"] = metadata.scripts
        ep["gui_scripts"] = metadata.gui_scripts
        for group, entries in ep.items():
            if entries:
                f.write(f"[{group}]\n")
                for name, target in entries.items():
                    f.write(f"{name} = {target}\n")
                f.write("\n")


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
    metadata_directory: str | None = None,
) -> str:
    settings_reader = SettingsReader(Path("pyproject.toml"), config_settings or {})
    settings = settings_reader.settings
    setup_logging(settings.logging.level)

    settings_reader.validate_may_exit()

    # We don't support preparing metadata yet
    assert metadata_directory is None

    with Path("pyproject.toml").open("rb") as ft:
        pyproject = tomllib.load(ft)
    metadata = StandardMetadata.from_pyproject(pyproject)

    if metadata.version is None:
        msg = "project.version is not statically specified, must be present currently"
        raise AssertionError(msg)

    wheel = DistWheel()
    wheel.dirname = wheel_directory
    wheel.name = packaging.utils.canonicalize_name(metadata.name).replace("-", "_")
    wheel.version = str(metadata.version)

    cmake = CMake.default_search(
        minimum_version=Version(settings.cmake.minimum_version)
    )
    rich_print(
        f"[green]***[/green] [bold][green]scikit-build-core {__version__}[/green] using [blue]CMake {cmake.version}[/blue]"
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        build_tmp_folder = Path(tmpdir)
        wheel_dir = build_tmp_folder / "wheel"
        build_dir = build_tmp_folder / "build"

        wheel_dirs = {
            "platlib": wheel_dir / "platlib",
            "data": wheel_dir / "data",
            "headers": wheel_dir / "headers",
            "scripts": wheel_dir / "scripts",
        }

        for wheel_dir in wheel_dirs.values():
            wheel_dir.mkdir(parents=True)

        if ".." in settings.wheel.install_dir:
            raise AssertionError("wheel.install_dir must not contain '..'")
        if settings.wheel.install_dir.startswith("/"):
            if not settings.experimental:
                raise AssertionError(
                    "Experimental features must be enabled to use absolute paths in wheel.install_dir"
                )
            if not settings.wheel.install_dir[1:].startswith(tuple(wheel_dirs)):
                raise AssertionError("Must target a valid wheel directory")
            install_dir = wheel_dir / settings.wheel.install_dir[1:]
        else:
            install_dir = wheel_dirs["platlib"] / settings.wheel.install_dir

        config = CMaker(
            cmake,
            source_dir=Path("."),
            build_dir=build_dir,
            build_type=settings.cmake.build_type,
        )

        builder = Builder(
            settings=settings,
            config=config,
        )

        rich_print("[green]***[/green] [bold]Configurating CMake...")
        defines: dict[str, str] = {}
        cache_entries = {f"SKBUILD_{k.upper()}_DIR": v for k, v in wheel_dirs.items()}
        builder.configure(
            defines=defines,
            cache_entries=cache_entries,
            name=metadata.name,
            version=metadata.version,
        )

        generator = builder.config.env.get(
            "CMAKE_GENERATOR", "MSVC" if sys.platform.startswith("win32") else "Unknown"
        )
        rich_print(
            f"[green]***[/green] [bold]Building project with [blue]{generator}[/blue]..."
        )
        build_args: list[str] = []
        builder.build(build_args=build_args)

        rich_print("[green]***[/green] [bold]Installing project into wheel...")
        builder.install(install_dir)

        rich_print("[green]***[/green] [bold]Making wheel...")
        _copy_python_packages_to_wheel(
            packages=settings.wheel.packages,
            name=metadata.name.replace("-", "_").replace(".", "_"),
            platlib_dir=wheel_dirs["platlib"],
            include=settings.sdist.include,
            exclude=settings.sdist.exclude,
        )

        for item in wheel_dirs["scripts"].iterdir():
            with item.open("rb") as f:
                content = f.read(len(b"#!python"))
                if content.startswith(b"#!/") and not b"#!python" != content:
                    logger.warning(
                        "Files in scripts/ are not post-processed yet for shabang fixes"
                    )
                    break

        _write_wheel_metadata(packages_dir=wheel_dirs["platlib"], metadata=metadata)

        tags = WheelTag.compute_best(
            builder.get_arch_tags(),
            settings.wheel.py_api,
            expand_macos=settings.wheel.expand_macos_universal_tags,
        )
        wheel.build({k: str(v) for k, v in wheel_dirs.items()}, tags=tags.tags_dict())

    wheel_filename: str = wheel.filename
    return wheel_filename



from __future__ import annotations

import dataclasses
import re
import sys
import sysconfig
from pathlib import Path
from typing import Mapping

from packaging.version import Version

from .. import __version__
from ..builder.sysconfig import get_python_include_dir, get_python_library
from ..cmake import CMaker
from ..errors import NinjaNotFoundError
from ..program_search import best_program, get_make_programs, get_ninja_programs
from ..resources import find_python
from ..settings.skbuild_model import ScikitBuildSettings

__all__: list[str] = ["Builder"]

DIR = Path(__file__).parent.resolve()


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass
class Builder:
    settings: ScikitBuildSettings
    config: CMaker

    # TODO: cross-compile support for other platforms
    def get_archs(self) -> list[str]:
        """
        Takes macOS platform settings and returns a list of platforms.

        Example (macOS):
            ARCHFLAGS="-arch x86_64" -> ["x86_64"]
            ARCHFLAGS="-arch x86_64 -arch arm64" -> ["x86_64", "arm64"]

        Returns an empty list otherwise or if ARCHFLAGS is not set.
        """

        if sys.platform.startswith("darwin"):
            archs = re.findall(r"-arch (\S+)", self.config.env.get("ARCHFLAGS", ""))
            return archs

        return []

    def get_arch_tags(self) -> list[str]:
        """
        This function returns tags suitable for use in wheels. The main
        difference between this method and get_archs() is that this returns
        universal2 instead of separate tags for x86_64 and arm64.
        """

        archs = self.get_archs()
        if sys.platform.startswith("darwin") and set(archs) == {"arm64", "x86_64"}:
            return ["universal2"]
        return archs

    def configure(
        self,
        *,
        defines: Mapping[str, str],
        cache_entries: Mapping[str, str | Path] | None = None,
        name: str | None = None,
        version: Version | None = None,
        limited_abi: bool | None = None,
    ) -> None:
        cmake_defines = dict(defines)
        cmake_args: list[str] = []

        site_packages = Path(sysconfig.get_path("purelib"))
        self.config.prefix_dirs.append(site_packages)
        if site_packages != DIR.parent.parent:
            self.config.prefix_dirs.append(DIR.parent.parent)

        fp_backport = self.settings.backport.find_python
        if fp_backport and self.config.cmake.version < Version(fp_backport):
            self.config.module_dirs.append(Path(find_python.__file__).parent.resolve())

        if sys.platform.startswith("win32"):
            # TODO: support cross-compilation
            is_64bit = sys.maxsize > 2**32
            if not is_64bit:
                cmake_args += ["-A", "Win32"]
        elif self.config.env.get(
            "CMAKE_GENERATOR", "Ninja"
        ) == "Ninja" and not self.config.env.get("CMAKE_MAKE_PROGRAM", ""):
            ninja = best_program(
                get_ninja_programs(),
                minimum_version=Version(self.settings.ninja.minimum_version),
            )
            if ninja is None:
                msg = "Ninja or make is required to build"
                if not self.settings.ninja.make_fallback:
                    raise NinjaNotFoundError(msg)

                make_programs = list(get_make_programs())
                if not make_programs:
                    raise NinjaNotFoundError(msg)
                self.config.env.setdefault("CMAKE_GENERATOR", "Unix Makefiles")
                self.config.env.setdefault("CMAKE_MAKE_PROGRAM", str(make_programs[0]))
            else:
                self.config.env.setdefault("CMAKE_GENERATOR", "Ninja")
                self.config.env.setdefault("CMAKE_MAKE_PROGRAM", str(ninja.path))

        cache_config: dict[str, str | Path] = {
            "SKBUILD": "2",
            "SKBUILD_CORE_VERSION": __version__,
        }

        if name is not None:
            canonical_name = name.replace("-", "_").replace(".", "_")
            cache_config["SKBUILD_PROJECT_NAME"] = canonical_name
        if version is not None:
            cache_config["SKBUILD_PROJECT_VERSION"] = str(version)

        # Classic Find Python
        python_library = get_python_library()
        python_include_dir = get_python_include_dir()
        cache_config["PYTHON_EXECUTABLE"] = sys.executable
        cache_config["PYTHON_INCLUDE_DIR"] = python_include_dir
        if python_library:
            cache_config["PYTHON_LIBRARY"] = python_library

        # Modern Find Python
        for prefix in ["Python", "Python3"]:
            cache_config[f"{prefix}_EXECUTABLE"] = sys.executable
            cache_config[f"{prefix}_ROOT_DIR"] = sys.prefix
            cache_config[f"{prefix}_INCLUDE_DIR"] = python_include_dir
            cache_config[f"{prefix}_FIND_REGISTRY"] = "NEVER"

        if limited_abi is None:
            limited_abi = self.settings.wheel.py_api.startswith("cp3")

        if limited_abi:
            cache_config["SKBUILD_SOABI"] = (
                "" if sys.platform.startswith("win") else "abi3"
            )
        else:
            # Workaround for bug in PyPy and packaging that is not handled in CMake
            # According to PEP 3149, SOABI and EXT_SUFFIX are interchangeable (and
            # the latter is much more likely to be correct as it is used elsewhere)
            ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
            assert ext_suffix
            cache_config["SKBUILD_SOABI"] = ext_suffix.rsplit(".", 1)[0].lstrip(".")

        if cache_entries:
            cache_config.update(cache_entries)

        self.config.init_cache(cache_config)

        # Add the pre-defined or passed CMake args
        cmake_args += self.settings.cmake.args

        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSX on conda-forge)
        cmake_args += [
            item for item in self.config.env.get("CMAKE_ARGS", "").split(" ") if item
        ]

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = self.get_archs()
            if archs:
                cmake_defines["CMAKE_OSX_ARCHITECTURES"] = ";".join(archs)

        # Add the pre-defined or passed CMake defines
        cmake_defines.update(self.settings.cmake.define)

        self.config.configure(
            defines=cmake_defines,
            cmake_args=cmake_args,
        )

    def build(self, build_args: list[str]) -> None:
        self.config.build(build_args=build_args, verbose=self.settings.cmake.verbose)

    def install(self, install_dir: Path) -> None:
        self.config.install(install_dir)



from __future__ import annotations

import dataclasses
import os
import subprocess
import sys
import textwrap
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Generator, TypeVar

from packaging.version import Version

from ._logging import logger
from ._shutil import Run
from .errors import CMakeConfigError, CMakeNotFoundError, FailedLiveProcessError
from .program_search import best_program, get_cmake_programs

__all__ = ["CMake", "CMaker"]


def __dir__() -> list[str]:
    return __all__


DIR = Path(__file__).parent.resolve()

Self = TypeVar("Self", bound="CMake")


@dataclasses.dataclass(frozen=True)
class CMake:
    version: Version
    cmake_path: Path

    @classmethod
    def default_search(
        cls: type[Self], *, minimum_version: Version | None = None, module: bool = True
    ) -> Self:
        candidates = get_cmake_programs(module=module)
        cmake_program = best_program(candidates, minimum_version=minimum_version)

        if cmake_program is None:
            raise CMakeNotFoundError(
                f"Could not find CMake with version >= {minimum_version}"
            )
        if cmake_program.version is None:
            msg = "CMake version undetermined @ {program.path}"
            raise CMakeNotFoundError(msg)

        return cls(version=cmake_program.version, cmake_path=cmake_program.path)

    def __fspath__(self) -> str:
        return os.fspath(self.cmake_path)


@dataclasses.dataclass
class CMaker:
    cmake: CMake
    source_dir: Path
    build_dir: Path
    build_type: str
    module_dirs: list[Path] = dataclasses.field(default_factory=list)
    prefix_dirs: list[Path] = dataclasses.field(default_factory=list)
    init_cache_file: Path = dataclasses.field(init=False, default=Path())
    env: dict[str, str] = dataclasses.field(init=False, default_factory=os.environ.copy)
    single_config: bool = not sys.platform.startswith("win32")

    def __post_init__(self) -> None:
        self.init_cache_file = self.build_dir / "CMakeInit.txt"

        if not self.source_dir.is_dir():
            raise CMakeConfigError(f"source directory {self.source_dir} does not exist")

        self.build_dir.mkdir(parents=True, exist_ok=True)
        if not self.build_dir.is_dir():
            raise CMakeConfigError(
                f"build directory {self.build} must be a (creatable) directory"
            )

    def init_cache(
        self, cache_settings: Mapping[str, str | os.PathLike[str] | bool]
    ) -> None:
        with self.init_cache_file.open("w", encoding="utf-8") as f:
            for key, value in cache_settings.items():
                if isinstance(value, bool):
                    value = "ON" if value else "OFF"
                    f.write(f'set({key} {value} CACHE BOOL "")\n')
                elif isinstance(value, os.PathLike):
                    # Convert to CMake's internal path format
                    value = str(value).replace("\\", "/")
                    f.write(f'set({key} [===[{value}]===] CACHE PATH "")\n')
                else:
                    f.write(f'set({key} [===[{value}]===] CACHE STRING "")\n')
        contents = self.init_cache_file.read_text(encoding="utf-8").strip()
        logger.debug(
            "{}:\n{}",
            self.init_cache_file,
            textwrap.indent(contents.strip(), "  "),
        )

    def _compute_cmake_args(
        self, settings: Mapping[str, str | os.PathLike[str] | bool]
    ) -> Generator[str, None, None]:
        yield f"-S{self.source_dir}"
        yield f"-B{self.build_dir}"

        if self.init_cache_file.is_file():
            yield f"-C{self.init_cache_file}"

        if self.single_config and self.build_type:
            yield f"-DCMAKE_BUILD_TYPE={self.build_type}"

        for key, value in settings.items():
            if isinstance(value, bool):
                value = "ON" if value else "OFF"
                yield f"-D{key}:BOOL={value}"
            elif isinstance(value, os.PathLike):
                value = str(value).replace("\\", "/")
                yield f"-D{key}:PATH={value}"
            else:
                yield f"-D{key}={value}"

        if self.module_dirs:
            # Convert to CMake's internal path format, otherwise this breaks try_compile on Windows
            module_dirs_str = ";".join(map(str, self.module_dirs)).replace("\\", "/")
            yield f"-DCMAKE_MODULE_PATH:PATH={module_dirs_str}"

        if self.prefix_dirs:
            prefix_dirs_str = ";".join(map(str, self.prefix_dirs)).replace("\\", "/")
            yield f"-DCMAKE_PREFIX_PATH:PATH={prefix_dirs_str}"

    def configure(
        self,
        *,
        defines: Mapping[str, str | os.PathLike[str] | bool] | None = None,
        cmake_args: Sequence[str] = (),
    ) -> None:
        if "CMAKE_GENERATOR" in self.env:
            gen = self.env["CMAKE_GENERATOR"]
            self.single_config = gen == "Ninja" or "Makefiles" in gen

        _cmake_args = self._compute_cmake_args(defines or {})

        try:
            Run(env=self.env).live(self.cmake, *_cmake_args, *cmake_args)
        except subprocess.CalledProcessError:
            msg = "CMake configuration failed"
            raise FailedLiveProcessError(msg) from None

    def _compute_build_args(
        self,
        *,
        verbose: bool,
    ) -> Generator[str, None, None]:
        if verbose:
            yield "-v"
        if self.build_type and not self.single_config:
            yield "--config"
            yield self.build_type

    def build(self, build_args: Sequence[str] = (), *, verbose: bool = False) -> None:
        local_args = self._compute_build_args(verbose=verbose)

        try:
            Run(env=self.env).live(
                self.cmake, "--build", self.build_dir, *build_args, *local_args
            )
        except subprocess.CalledProcessError:
            msg = "CMake build failed"
            raise FailedLiveProcessError(msg) from None

    def install(self, prefix: Path) -> None:
        opts: list[str] = []
        if not self.single_config and self.build_type:
            opts += ["--config", self.build_type]
        try:
            Run(env=self.env).live(
                self.cmake,
                "--install",
                self.build_dir,
                "--prefix",
                prefix,
                *opts,
            )
        except subprocess.CalledProcessError:
            msg = "CMake install failed"
            raise FailedLiveProcessError(msg) from None



from __future__ import annotations

import os
import shutil
import sys
from pathlib import Path

from scikit_build_core.builder.get_requires import cmake_ninja_for_build_wheel

ninja = [] if sys.platform.startswith("win") else ["ninja>=1.5"]


def which_mock(name: str) -> str | None:
    if name in ("ninja", "ninja-build", "cmake3", "samu", "gmake", "make"):
        return None
    if name == "cmake":
        return "cmake/path"
    return None


def test_get_requires_for_build_wheel(fp, monkeypatch):
    cmake = Path("cmake/path").resolve()
    monkeypatch.setattr(shutil, "which", which_mock)
    fp.register([os.fspath(cmake), "--version"], stdout="3.14.0")
    assert cmake_ninja_for_build_wheel() == ["cmake>=3.15", *ninja]


def test_get_requires_for_build_wheel_uneeded(fp, monkeypatch):
    cmake = Path("cmake/path").resolve()
    monkeypatch.setattr(shutil, "which", which_mock)
    fp.register([os.fspath(cmake), "--version"], stdout="3.18.0")
    assert cmake_ninja_for_build_wheel() == [*ninja]


def test_get_requires_for_build_wheel_settings(fp, monkeypatch):
    cmake = Path("cmake/path").resolve()
    monkeypatch.setattr(shutil, "which", which_mock)
    fp.register([os.fspath(cmake), "--version"], stdout="3.18.0")
    config = {"cmake.minimum-version": "3.20"}
    assert cmake_ninja_for_build_wheel(config) == [
        "cmake>=3.20",
        *ninja,
    ]


def test_get_requires_for_build_wheel_pyproject(fp, monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    tmp_path.joinpath("pyproject.toml").write_text(
        """
        [tool.scikit-build.cmake]
        minimum-version = "3.21"
        """
    )
    cmake = Path("cmake/path").resolve()
    monkeypatch.setattr(shutil, "which", which_mock)
    fp.register([os.fspath(cmake), "--version"], stdout="3.18.0")
    assert cmake_ninja_for_build_wheel() == ["cmake>=3.21", *ninja]



from __future__ import annotations

import os
import shutil
import sys
import sysconfig
from pathlib import Path

import setuptools
import setuptools.command.build_ext
from packaging.version import Version
from setuptools.dist import Distribution

from .._compat.typing import Literal
from ..builder.builder import Builder
from ..builder.macos import normalize_macos_version
from ..cmake import CMake, CMaker
from ..settings.skbuild_read_settings import SettingsReader

__all__: list[str] = ["CMakeExtension", "cmake_extensions"]


def __dir__() -> list[str]:
    return __all__


# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# The sourcedir is relative to the setup.py directory, where the CMakeLists.txt lives
class CMakeExtension(setuptools.Extension):
    def __init__(self, name: str, sourcedir: str = "", **kwargs: object) -> None:
        super().__init__(name, [], **kwargs)
        self.sourcedir = Path(sourcedir).resolve()


class CMakeBuild(setuptools.command.build_ext.build_ext):
    def build_extension(self, ext: setuptools.Extension) -> None:
        if not isinstance(ext, CMakeExtension):
            super().build_extension(ext)
            return

        build_tmp_folder = Path(self.build_temp)
        build_temp = build_tmp_folder / "_skbuild"  # TODO: include python platform

        dist = self.distribution  # type: ignore[attr-defined]

        limited_api = dist.get_command_obj("bdist_wheel").py_limited_api
        if limited_api:
            ext.py_limited_api = True

        # This dir doesn't exist, so Path.cwd() is needed for Python < 3.10
        # due to a Windows bug in resolve https://github.com/python/cpython/issues/82852
        ext_fullpath = Path.cwd() / self.get_ext_fullpath(ext.name)  # type: ignore[no-untyped-call]
        extdir = ext_fullpath.parent.resolve()

        # TODO: this is a hack due to moving temporary paths for isolation
        if build_temp.exists():
            shutil.rmtree(build_temp)

        settings = SettingsReader(Path("pyproject.toml"), {}).settings

        cmake = CMake.default_search(
            minimum_version=Version(settings.cmake.minimum_version)
        )

        config = CMaker(
            cmake,
            source_dir=ext.sourcedir,
            build_dir=build_temp,
            build_type=settings.cmake.build_type,
        )

        builder = Builder(
            settings=settings,
            config=config,
        )

        # Setuptools requires this be specified if there's a mismatch.
        if sys.platform.startswith("darwin"):
            arm_only = builder.get_archs() == ["arm64"]
            orig_macos_str = sysconfig.get_platform().rsplit("-", 1)[0].split("-", 1)[1]
            orig_macos = normalize_macos_version(orig_macos_str, arm_only)
            config.env.setdefault("MACOSX_DEPLOYMENT_TARGET", str(orig_macos))

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        builder.config.build_type = "Debug" if debug else "Release"

        defines: dict[str, str] = {}

        for key, value in ext.define_macros:
            assert isinstance(value, str), "define_macros values must not be None"
            defines[key] = value

        builder.configure(
            defines=defines,
            name=dist.get_name(),
            version=dist.get_version(),
            limited_abi=ext.py_limited_api,
        )

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        build_args = []
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in builder.config.env:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                build_args.append(f"-j{self.parallel}")

        builder.build(build_args=build_args)
        builder.install(extdir)


def cmake_extensions(
    dist: Distribution, attr: Literal["cmake_extensions"], value: list[CMakeExtension]
) -> None:
    settings = SettingsReader(Path("pyproject.toml"), {}).settings

    assert attr == "cmake_extensions"
    assert len(value) > 0

    assert (
        not settings.wheel.expand_macos_universal_tags
    ), "wheel.expand_macos_universal_tags is not supported in setuptools mode"
    assert (
        settings.logging.level == "WARNING"
    ), "Logging is not adjustable in setuptools mode yet"
    assert (
        not settings.wheel.py_api
    ), "wheel.py_api is not supported in setuptools mode, use bdist_wheel options instead"

    dist.has_ext_modules = lambda: True  # type: ignore[attr-defined]
    dist.ext_modules = (dist.ext_modules or []) + value

    dist.cmdclass["build_ext"] = CMakeBuild


def cmake_source_dir(
    dist: Distribution, attr: Literal["cmake_source_dir"], value: str
) -> None:
    assert attr == "cmake_source_dir"
    assert Path(value).is_dir()
    assert dist.cmake_extensions is None, "Combining cmake_source_dir= and cmake_extensions= is not allowed"  # type: ignore[attr-defined]
    name = dist.get_name().replace("-", "_")  # type: ignore[attr-defined]

    extensions = [CMakeExtension(name, value)]
    dist.cmake_extensions = extensions  # type: ignore[attr-defined]
    cmake_extensions(dist, "cmake_extensions", extensions)



# pylint: disable=unused-argument

from __future__ import annotations

import sys

__all__ = [
    "build_sdist",
    "build_wheel",
    "get_requires_for_build_sdist",
    "get_requires_for_build_wheel",
]


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
    metadata_directory: str | None = None,
) -> str:
    from ..errors import FailedLiveProcessError
    from .._logging import rich_print
    from .wheel import build_wheel as skbuild_build_wheel

    try:
        return skbuild_build_wheel(wheel_directory, config_settings, metadata_directory)
    except FailedLiveProcessError as err:
        sys.stdout.flush()
        rich_print(f"\n[red bold]*** {' '.join(err.args)}", file=sys.stderr)
        raise SystemExit(1) from None


def build_sdist(
    sdist_directory: str,
    config_settings: dict[str, list[str] | str] | None = None,
) -> str:
    from .sdist import build_sdist as skbuild_build_sdist

    return skbuild_build_sdist(sdist_directory, config_settings)


def get_requires_for_build_sdist(
    # pylint: disable-next=unused-argument
    config_settings: dict[str, str | list[str]]
    | None = None
) -> list[str]:
    return ["pathspec", "pyproject_metadata"]


def get_requires_for_build_wheel(
    config_settings: dict[str, str | list[str]] | None = None,
) -> list[str]:
    from ..builder.get_requires import cmake_ninja_for_build_wheel

    return ["distlib", "pathspec", "pyproject_metadata"] + cmake_ninja_for_build_wheel(
        config_settings
    )



from __future__ import annotations

import dataclasses
import os
from collections.abc import Generator, Mapping, Sequence
from typing import Any, TypeVar, Union

from .._compat.builtins import ExceptionGroup
from .._compat.typing import Protocol, runtime_checkable

T = TypeVar("T")

__all__ = ["Source", "SourceChain", "ConfSource", "EnvSource", "TOMLSource"]


def __dir__() -> list[str]:
    return __all__


def _dig_strict(dict_: Mapping[str, Any], *names: str) -> Any:
    for name in names:
        dict_ = dict_[name]
    return dict_


def _dig_not_strict(dict_: Mapping[str, Any], *names: str) -> Any:
    for name in names:
        dict_ = dict_.get(name, {})
    return dict_


def _dig_fields(__opt: object, *names: str) -> Any:
    for name in names:
        fields = dataclasses.fields(__opt)
        types = [x.type for x in fields if x.name == name]
        if len(types) != 1:
            raise KeyError("Could not access {'.'.join(names)}")
        (__opt,) = types
    return __opt


@runtime_checkable
class TypeLike(Protocol):
    @property
    def __origin__(self) -> Any:
        ...

    @property
    def __args__(self) -> list[Any]:
        ...


def _process_union(target: type[Any]) -> Any:
    """
    Selects the non-None item in an Optional or Optional-like Union. Passes through non-Unions.
    """

    if (
        not isinstance(target, TypeLike)
        or not hasattr(target, "__origin__")
        or target.__origin__ is not Union
    ):
        return target

    if len(target.__args__) == 2:
        items = list(target.__args__)
        if type(None) not in items:
            raise AssertionError(f"None must be in union, got {items}")
        items.remove(type(None))
        return items[0]

    raise AssertionError("Only Unions with None supported")


def _get_target_raw_type(target: type[Any]) -> type[Any]:
    """
    Takes a type like Optional[str] and returns str,
    or Optional[Dict[str, int]] and returns dict.
    """

    target = _process_union(target)
    # The hasattr is required for Python 3.7, though not quite sure why
    if isinstance(target, TypeLike) and hasattr(target, "__origin__"):
        return target.__origin__
    return target


def _get_inner_type(target: type[Any]) -> type[Any]:
    """
    Takes a types like List[str] and returns str,
    or Dict[str, int] and returns int.
    """

    raw_target = _get_target_raw_type(target)
    target = _process_union(target)
    if raw_target == list:
        assert isinstance(target, TypeLike)
        return target.__args__[0]
    if raw_target == dict:
        assert isinstance(target, TypeLike)
        return target.__args__[1]
    raise AssertionError("Expected a list or dict")


class Source(Protocol):
    def has_item(self, *fields: str, is_dict: bool) -> bool:
        """
        Check if the source contains a chain of fields. For example, feilds =
        [Field(name="a"), Field(name="b")] will check if the source contains the
        key "a.b".
        """
        ...

    def get_item(self, *fields: str, is_dict: bool) -> Any:
        ...

    @classmethod
    def convert(cls, item: Any, target: type[Any]) -> object:
        ...

    def unrecognized_options(self, options: object) -> Generator[str, None, None]:
        ...


class EnvSource:
    """
    This is a source using environment variables.
    """

    def __init__(self, prefix: str, *, env: Mapping[str, str] | None = None) -> None:
        self.env = env or os.environ
        self.prefix = prefix

    def _get_name(self, *fields: str) -> str:
        names = [field.upper() for field in fields]
        return "_".join([self.prefix, *names] if self.prefix else names)

    # pylint: disable-next=unused-argument
    def has_item(self, *fields: str, is_dict: bool) -> bool:
        name = self._get_name(*fields)
        return name in self.env and self.env[name] != ""

    # pylint: disable-next=unused-argument
    def get_item(self, *fields: str, is_dict: bool) -> str | dict[str, str]:
        name = self._get_name(*fields)
        if name in self.env:
            return self.env[name]
        raise KeyError(f"{name!r} not found in environment")

    @classmethod
    def convert(cls, item: str, target: type[Any]) -> object:
        raw_target = _get_target_raw_type(target)
        if raw_target == list:
            return [
                cls.convert(i.strip(), _get_inner_type(target)) for i in item.split(";")
            ]
        if raw_target == dict:
            items = (i.strip().split("=") for i in item.split(";"))
            return {k: cls.convert(v, _get_inner_type(target)) for k, v in items}

        if raw_target is bool:
            result = item.strip().lower() not in {"0", "false", "off", "no", ""}
            return result

        if callable(raw_target):
            return raw_target(item)
        raise AssertionError(f"Can't convert target {target}")

    # pylint: disable-next=unused-argument
    def unrecognized_options(self, options: object) -> Generator[str, None, None]:
        yield from ()


def _unrecognized_dict(
    settings: Mapping[str, Any], options: object, above: Sequence[str]
) -> Generator[str, None, None]:
    for keystr in settings:
        matches = [
            x for x in dataclasses.fields(options) if x.name.replace("_", "-") == keystr
        ]
        if not matches:
            yield ".".join((*above, keystr))
            continue
        (inner_option_field,) = matches
        inner_option = inner_option_field.type
        if dataclasses.is_dataclass(inner_option):
            yield from _unrecognized_dict(
                settings[keystr], inner_option, (*above, keystr)
            )


class ConfSource:
    """
    This is a source for the PEP 517 configuration settings.
    You should initialize it with a dict from PEP 517. a.b will be treated as
    nested dicts. "verify" is a boolean that determines whether unrecognized
    options should be checked for. Only set this to false if this might be sharing
    config options at the same level.
    """

    def __init__(
        self,
        *prefixes: str,
        settings: Mapping[str, str | list[str]],
        verify: bool = True,
    ):
        self.prefixes = prefixes
        self.settings = settings
        self.verify = verify

    def _get_name(self, *fields: str) -> list[str]:
        names = [field.replace("_", "-") for field in fields]
        return [*self.prefixes, *names]

    def has_item(self, *fields: str, is_dict: bool) -> bool:
        names = self._get_name(*fields)
        name = ".".join(names)

        if is_dict:
            return any(k.startswith(f"{name}.") for k in self.settings)

        return name in self.settings

    def get_item(self, *fields: str, is_dict: bool) -> str | list[str] | dict[str, str]:
        names = self._get_name(*fields)
        name = ".".join(names)
        if is_dict:
            d = {
                k[len(name) + 1 :]: str(v)
                for k, v in self.settings.items()
                if k.startswith(f"{name}.")
            }
            if d:
                return d
            raise KeyError(f"Dict items {name}.* not found in settings")
        if name in self.settings:
            return self.settings[name]

        raise KeyError(f"{name!r} not found in configuration settings")

    @classmethod
    def convert(
        cls, item: str | list[str] | dict[str, str], target: type[Any]
    ) -> object:
        raw_target = _get_target_raw_type(target)
        if raw_target == list:
            if isinstance(item, list):
                return [cls.convert(i, _get_inner_type(target)) for i in item]
            assert not isinstance(item, dict)
            return [
                cls.convert(i.strip(), _get_inner_type(target)) for i in item.split(";")
            ]
        if raw_target == dict:
            assert not isinstance(item, (str, list))
            return {k: cls.convert(v, _get_inner_type(target)) for k, v in item.items()}
        assert not isinstance(
            item, (list, dict)
        ), "Can't convert list or dict to non-list/dict"
        if raw_target is bool:
            result = item.strip().lower() not in {"0", "false", "off", "no", ""}
            return result
        if callable(raw_target):
            return raw_target(item)
        raise AssertionError(f"Can't convert target {target}")

    def unrecognized_options(self, options: object) -> Generator[str, None, None]:
        if not self.verify:
            return
        for keystr in self.settings:
            keys = keystr.replace("-", "_").split(".")[len(self.prefixes) :]
            try:
                outer_option = _dig_fields(options, *keys[:-1])
            except KeyError:
                yield ".".join(keystr.split(".")[:-1])
                continue
            if dataclasses.is_dataclass(outer_option):
                try:
                    _dig_fields(outer_option, keys[-1])
                except KeyError:
                    yield keystr
                    continue
            if _get_target_raw_type(outer_option) == dict:
                continue


class TOMLSource:
    def __init__(self, *prefixes: str, settings: Mapping[str, Any]):
        self.prefixes = prefixes
        self.settings = _dig_not_strict(settings, *prefixes)

    def _get_name(self, *fields: str) -> list[str]:
        return [field.replace("_", "-") for field in fields]

    # pylint: disable-next=unused-argument
    def has_item(self, *fields: str, is_dict: bool) -> bool:
        names = self._get_name(*fields)
        try:
            _dig_strict(self.settings, *names)
            return True
        except KeyError:
            return False

    # pylint: disable-next=unused-argument
    def get_item(self, *fields: str, is_dict: bool) -> Any:
        names = self._get_name(*fields)
        try:
            return _dig_strict(self.settings, *names)
        except KeyError:
            raise KeyError(f"{names!r} not found in configuration settings") from None

    @classmethod
    def convert(cls, item: Any, target: type[Any]) -> object:
        raw_target = _get_target_raw_type(target)
        if raw_target == list:
            return [cls.convert(it, _get_inner_type(target)) for it in item]
        if raw_target == dict:
            return {k: cls.convert(v, _get_inner_type(target)) for k, v in item.items()}
        if callable(raw_target):
            return raw_target(item)
        raise AssertionError(f"Can't convert target {target}")

    def unrecognized_options(self, options: object) -> Generator[str, None, None]:
        yield from _unrecognized_dict(self.settings, options, self.prefixes)


class SourceChain:
    def __init__(self, *sources: Source):
        self.sources = sources

    def has_item(self, *fields: str, is_dict: bool) -> bool:
        for source in self.sources:
            if source.has_item(*fields, is_dict=is_dict):
                return True
        return False

    def get_item(self, *fields: str, is_dict: bool) -> Any:
        for source in self.sources:
            if source.has_item(*fields, is_dict=is_dict):
                return source.get_item(*fields, is_dict=is_dict)
        raise KeyError(f"{fields!r} not found in any source")

    @classmethod
    def convert(cls, item: Any, target: type[T]) -> T:
        raise NotImplementedError(
            "SourceChain cannot convert items, use the result from has_item"
        )

    def convert_target(self, target: type[T], *prefixes: str) -> T:
        errors = []
        prep = {}
        for field in dataclasses.fields(target):
            if dataclasses.is_dataclass(field.type):
                try:
                    prep[field.name] = self.convert_target(
                        field.type, *prefixes, field.name
                    )
                except Exception as e:
                    errors.append(e)
                continue

            is_dict = _get_target_raw_type(field.type) == dict

            for source in self.sources:
                if source.has_item(*prefixes, field.name, is_dict=is_dict):
                    simple = source.get_item(*prefixes, field.name, is_dict=is_dict)
                    try:
                        tmp = source.convert(simple, field.type)
                    except Exception as e:
                        errors.append(e)
                        prep[field.name] = None
                        break

                    if is_dict:
                        assert isinstance(tmp, dict), f"{field.name} must be a dict"
                        prep[field.name] = {**tmp, **prep.get(field.name, {})}
                        continue

                    prep[field.name] = tmp
                    break

            if field.name in prep:
                continue

            if field.default is not dataclasses.MISSING:
                prep[field.name] = field.default
                continue
            if field.default_factory is not dataclasses.MISSING:
                prep[field.name] = field.default_factory()
                continue

            errors.append(ValueError(f"Missing value for {field.name!r}"))

        if errors:
            prefix_str = ".".join(prefixes)
            raise ExceptionGroup(f"Failed converting {prefix_str}", errors)

        return target(**prep)

    def unrecognized_options(self, options: object) -> Generator[str, None, None]:
        for source in self.sources:
            yield from source.unrecognized_options(options)



from __future__ import annotations

from ..builder.get_requires import cmake_ninja_for_build_wheel

__all__ = [
    "prepare_metadata_for_build_wheel",
    "build_wheel",
    "build_sdist",
    "get_requires_for_build_sdist",
    "get_requires_for_build_wheel",
]


def __dir__() -> list[str]:
    return __all__


def get_requires_for_build_sdist(
    # pylint: disable-next=unused-argument
    config_settings: dict[str, str | list[str]]
    | None = None
) -> list[str]:
    return ["setuptools"]


def get_requires_for_build_wheel(
    config_settings: dict[str, str | list[str]] | None = None
) -> list[str]:
    return ["setuptools", "wheel"] + cmake_ninja_for_build_wheel(config_settings)


def build_sdist(
    sdist_directory: str,
    config_settings: dict[str, str | list[str]] | None = None,
) -> str:
    import setuptools.build_meta

    return setuptools.build_meta.build_sdist(  # type: ignore[no-any-return]
        sdist_directory, config_settings
    )


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, str | list[str]] | None = None,
    metadata_directory: str | None = None,
) -> str:
    import setuptools.build_meta

    return setuptools.build_meta.build_wheel(  # type: ignore[no-any-return]
        wheel_directory, config_settings, metadata_directory
    )


def prepare_metadata_for_build_wheel(
    metadata_directory: str,
    config_settings: dict[str, str | list[str]] | None = None,
) -> str:
    import setuptools.build_meta

    return setuptools.build_meta.prepare_metadata_for_build_wheel(  # type: ignore[no-any-return]
        metadata_directory, config_settings
    )



from __future__ import annotations

import functools
import os
import sys
from collections.abc import Mapping
from pathlib import Path

from packaging.tags import sys_tags
from packaging.version import Version

from .._compat import tomllib
from .._compat.typing import Literal
from ..program_search import (
    best_program,
    get_cmake_programs,
    get_make_programs,
    get_ninja_programs,
)
from ..resources import resources
from ..settings.skbuild_read_settings import SettingsReader

__all__ = ["cmake_ninja_for_build_wheel"]


def __dir__() -> list[str]:
    return __all__


@functools.lru_cache(maxsize=2)
def known_wheels(name: Literal["ninja", "cmake"]) -> frozenset[str]:
    with resources.joinpath("known_wheels.toml").open("rb") as f:
        return frozenset(tomllib.load(f)["tool"]["scikit-build"][name]["known-wheels"])


@functools.lru_cache(maxsize=2)
def is_known_platform(platforms: frozenset[str]) -> bool:
    for tag in sys_tags():
        if tag.platform in platforms:
            return True
    return False


def cmake_ninja_for_build_wheel(
    # pylint: disable-next=unused-argument
    config_settings: Mapping[str, str | list[str]]
    | None = None
) -> list[str]:

    settings = SettingsReader(Path("pyproject.toml"), config_settings or {}).settings

    packages = []
    cmake_min = Version(settings.cmake.minimum_version)
    cmake = best_program(get_cmake_programs(module=False), minimum_version=cmake_min)
    if cmake is None:
        packages.append(f"cmake>={cmake_min}")

    if (
        not sys.platform.startswith("win")
        and os.environ.get("CMAKE_GENERATOR", "Ninja") == "Ninja"
        and not os.environ.get("CMAKE_MAKE_PROGRAM", "")
    ):
        ninja_min = Version(settings.ninja.minimum_version)
        ninja = best_program(
            get_ninja_programs(module=False), minimum_version=ninja_min
        )
        if ninja is None:
            if (
                not settings.ninja.make_fallback
                or is_known_platform(known_wheels("ninja"))
                or not list(get_make_programs())
            ):
                packages.append(f"ninja>={ninja_min}")

    return packages



from __future__ import annotations

import sys

if sys.version_info < (3, 8):
    from typing_extensions import Literal, Protocol, runtime_checkable
else:
    from typing import Literal, Protocol, runtime_checkable

__all__ = ["Protocol", "runtime_checkable", "Literal"]


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import logging
import re
import sys
import typing
from typing import Any

__all__ = ["logger", "raw_logger", "ScikitBuildLogger", "rich_print"]


def __dir__() -> list[str]:
    return __all__


raw_logger = logging.getLogger(
    "scikit_build_core"
)  # TODO: maybe should be scikit-build?
raw_logger.setLevel(logging.DEBUG)  # TODO: configure


class FStringMessage:
    "This class captures a formatted string message and only produces it on demand."

    def __init__(self, fmt: str, *args: object, **kwargs: object) -> None:
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return self.fmt.format(*self.args, **self.kwargs)

    def __repr__(self) -> str:
        return (
            f"<FStringMessage {self.fmt!r} args={self.args!r} kwargs={self.kwargs!r}>"
        )


if sys.version_info < (3, 8):
    opts: Any = {}
else:
    opts = {"stacklevel": 2}


class ScikitBuildLogger:
    # pylint: disable-next=redefined-outer-name
    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger

    def debug(self, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.debug(FStringMessage(msg, *args, **kwargs), **opts)

    def info(self, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.info(FStringMessage(msg, *args, **kwargs), **opts)

    def warning(self, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.warning(FStringMessage(msg, *args, **kwargs), **opts)

    def error(self, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.error(FStringMessage(msg, *args, **kwargs), **opts)

    def critical(self, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.critical(FStringMessage(msg, *args, **kwargs), **opts)

    def exception(self, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.exception(FStringMessage(msg, *args, **kwargs), **opts)

    def log(self, level: int, msg: str, *args: object, **kwargs: object) -> None:
        self.logger.log(level, FStringMessage(msg, *args, **kwargs), **opts)

    def setLevel(self, level: int) -> None:
        self.logger.setLevel(level)

    def addHandler(self, handler: logging.Handler) -> None:
        self.logger.addHandler(handler)


logger = ScikitBuildLogger(raw_logger)


def _strip_rich(msg: object) -> object:
    if isinstance(msg, str):
        return re.sub(r"\[.*?\]", "", msg)
    return msg


def fake_rich_print(*args: object, **kwargs: object) -> None:
    args_2 = (_strip_rich(arg) for arg in args)
    print(*args_2, **kwargs)  # type: ignore[call-overload] # noqa: T201


try:
    from rich import print as rich_print
except ModuleNotFoundError:
    if not typing.TYPE_CHECKING:
        rich_print = fake_rich_print



from __future__ import annotations

import sys
from collections.abc import Generator, Mapping
from pathlib import Path

from packaging.version import Version

from .. import __version__
from .._compat import tomllib
from .._logging import logger, rich_print
from ..errors import CMakeConfigError
from .skbuild_model import ScikitBuildSettings
from .sources import ConfSource, EnvSource, SourceChain, TOMLSource

__all__ = ["SettingsReader"]


def __dir__() -> list[str]:
    return __all__


class SettingsReader:
    def __init__(
        self,
        pyproject_toml: Path,
        config_settings: Mapping[str, str | list[str]],
        verify_conf: bool = True,
    ) -> None:
        with pyproject_toml.open("rb") as f:
            pyproject = tomllib.load(f)

        self.sources = SourceChain(
            EnvSource("SKBUILD"),
            ConfSource(settings=config_settings, verify=verify_conf),
            TOMLSource("tool", "scikit-build", settings=pyproject),
        )
        self.settings = self.sources.convert_target(ScikitBuildSettings)

        if self.settings.minimum_version:
            current_version = Version(__version__)
            minimum_version = Version(self.settings.minimum_version)
            if current_version < minimum_version:
                msg = (
                    f"scikit-build-core version {__version__} is too old. "
                    f"Minimum required version is {self.settings.minimum_version}."
                )
                raise CMakeConfigError(msg)

    def unrecognized_options(self) -> Generator[str, None, None]:
        return self.sources.unrecognized_options(ScikitBuildSettings)

    def validate_may_exit(self) -> None:
        unrecognized = list(self.unrecognized_options())
        if unrecognized:
            if self.settings.strict_config:
                sys.stdout.flush()
                rich_print(
                    "[red][bold]ERROR:[/bold] Unrecognized options:", file=sys.stderr
                )
                for option in unrecognized:
                    rich_print(f"  [red]{option}", file=sys.stderr)
                raise SystemExit(7)
            logger.warning("Unrecognized options: {}", ", ".join(unrecognized))



from __future__ import annotations

import os
import sysconfig
from pathlib import Path

from .._logging import logger

__all__ = ["get_python_include_dir", "get_python_library"]


def __dir__() -> list[str]:
    return __all__


def get_python_library() -> Path | None:
    libdirstr = sysconfig.get_config_var("LIBDIR")
    ldlibrarystr = sysconfig.get_config_var("LDLIBRARY")
    libdir: Path | None = libdirstr and Path(libdirstr)
    ldlibrary: Path | None = ldlibrarystr and Path(ldlibrarystr)
    multiarch: str | None = sysconfig.get_config_var("MULTIARCH")
    masd: str | None = sysconfig.get_config_var("multiarchsubdir")

    if libdir and ldlibrary:
        try:
            libdir_is_dir = libdir.is_dir()
        except PermissionError:
            return None
        if libdir_is_dir:
            if multiarch and masd:
                if masd.startswith(os.sep):
                    masd = masd[len(os.sep) :]
                libdir_masd = libdir / masd
                if libdir_masd.is_dir():
                    libdir = libdir_masd
            libpath = libdir / ldlibrary
            if Path(os.path.expandvars(libpath)).is_file():
                return libpath

    framework_prefix = sysconfig.get_config_var("PYTHONFRAMEWORKPREFIX")
    if framework_prefix and Path(framework_prefix).is_dir() and ldlibrary:
        libpath = Path(framework_prefix) / ldlibrary
        if libpath.is_file():
            return libpath

    logger.warning(
        "Can't find a Python library, got libdir={}, ldlibrary={}, multiarch={}, masd={}",
        libdir,
        ldlibrary,
        multiarch,
        masd,
    )
    return None


def get_python_include_dir() -> Path:
    return Path(sysconfig.get_path("include"))



import hashlib
import shutil
import subprocess
import sys
import tarfile
import textwrap
import zipfile
from pathlib import Path

import pytest

DIR = Path(__file__).parent.resolve()
HELLO_PEP518 = DIR / "packages/simple_pyproject_ext"


@pytest.mark.integration
def test_pep518_sdist():
    correct_metadata = textwrap.dedent(
        """\
        Metadata-Version: 2.1
        Name: cmake-example
        Version: 0.0.1
        Requires-Python: >=3.7
        Provides-Extra: test
        Requires-Dist: pytest>=6.0; extra == "test"
        """
    )

    dist = HELLO_PEP518 / "dist"
    shutil.rmtree(dist, ignore_errors=True)
    subprocess.run(
        [sys.executable, "-m", "build", "--sdist"], cwd=HELLO_PEP518, check=True
    )
    (sdist,) = dist.iterdir()
    assert "cmake-example-0.0.1.tar.gz" == sdist.name

    if not (sys.platform.startswith("win32") or sys.platform.startswith("cygwin")):
        hash = hashlib.sha256(sdist.read_bytes()).hexdigest()
        if sys.version_info < (3, 9):
            assert (
                hash
                == "600ed996e51642027557759ee9eeb31b5cae1f443313f5f7d0a40d9cc9cbdd13"
            )
        else:
            assert (
                hash
                == "4f47a4e797db1cb8e15afb368360d5f2ac5ae4b6c7e38e0771f8eba65fab65e4"
            )

    with tarfile.open(sdist) as f:
        file_names = set(f.getnames())
        assert file_names == {
            f"cmake-example-0.0.1/{x}"
            for x in (
                "CMakeLists.txt",
                "pyproject.toml",
                "src/main.cpp",
                "PKG-INFO",
            )
        }
        pkg_info = f.extractfile("cmake-example-0.0.1/PKG-INFO")
        assert pkg_info
        pkg_info_contents = pkg_info.read().decode()
        assert correct_metadata == pkg_info_contents


@pytest.mark.compile
@pytest.mark.configure
@pytest.mark.integration
@pytest.mark.parametrize(
    "build_args", [(), ("--wheel",)], ids=["sdist_to_wheel", "wheel_directly"]
)
def test_pep518_wheel(virtualenv, build_args, monkeypatch):
    dist = HELLO_PEP518 / "dist"
    shutil.rmtree(dist, ignore_errors=True)
    monkeypatch.chdir(HELLO_PEP518)
    virtualenv.install("build[virtualenv]")
    virtualenv.module(
        "build",
        "--config-setting=logging.level=DEBUG",
        *build_args,
    )
    (wheel,) = dist.glob("cmake_example-0.0.1-*.whl")

    if sys.version_info >= (3, 8):
        with wheel.open("rb") as f:
            p = zipfile.Path(f)
            file_names = [p.name for p in p.iterdir()]

        assert len(file_names) == 2
        assert "cmake_example-0.0.1.dist-info" in file_names
        file_names.remove("cmake_example-0.0.1.dist-info")
        (so_file,) = file_names

        assert so_file.startswith("cmake_example")
        print("SOFILE:", so_file)

    virtualenv.install(wheel)

    version = virtualenv.execute(
        "import cmake_example; print(cmake_example.__version__)"
    )
    assert version == "0.0.1"

    add = virtualenv.execute("import cmake_example; print(cmake_example.add(1, 2))")
    assert add == "3"


@pytest.mark.compile
@pytest.mark.configure
@pytest.mark.integration
def test_pep518_pip(virtualenv):
    virtualenv.install("-v", HELLO_PEP518)

    version = virtualenv.execute(
        "import cmake_example; print(cmake_example.__version__)"
    )
    assert version == "0.0.1"

    add = virtualenv.execute("import cmake_example; print(cmake_example.add(1, 2))")
    assert add == "3"



import shutil
import sys
import zipfile
from pathlib import Path

import pytest

from scikit_build_core.setuptools.build_meta import build_wheel

pytestmark = pytest.mark.setuptools

DIR = Path(__file__).parent.resolve()
ABI_PKG = DIR / "packages/abi3_setuptools_ext"


@pytest.mark.compile
@pytest.mark.configure
@pytest.mark.skipif(
    sys.implementation.name == "pypy", reason="pypy does not support abi3"
)
@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="abi3 is hard to target with FindPython on Windows",
)
def test_abi3_wheel(tmp_path, monkeypatch, virtualenv):
    dist = tmp_path / "dist"
    dist.mkdir()
    monkeypatch.chdir(ABI_PKG)
    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_wheel(str(dist))
    (wheel,) = dist.glob("abi3_example-0.0.1-*.whl")
    assert wheel == dist / out
    assert "-cp37-abi3-" in out

    assert virtualenv.execute("print('hello')") == "hello"

    if sys.version_info >= (3, 8):
        with wheel.open("rb") as f:
            p = zipfile.Path(f)
            file_names = {p.name for p in p.iterdir()}

        so_file = (
            "abi3_example.abi3.dll"
            if sys.platform.startswith("cygwin")
            else "abi3_example.abi3.so"
        )
        assert so_file in file_names

    virtualenv.install(wheel)

    output = virtualenv.execute("import abi3_example; print(abi3_example.square(2))")
    assert output == "4.0"



from __future__ import annotations

import functools
import logging

from .._logging import logger

__all__ = ["setup_logging"]


def __dir__() -> list[str]:
    return __all__


@functools.lru_cache(1)
def setup_logging(log_level: str) -> None:
    level_value = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }[log_level]

    try:
        import rich.logging
        import rich.traceback
    except ModuleNotFoundError:
        ch = logging.StreamHandler()
        ch.setLevel(level_value)
        # create formatter and add it to the handlers
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(ch)
        return

    FORMAT = "%(message)s"
    logging.basicConfig(
        level=level_value,
        format=FORMAT,
        datefmt="[%X]",
        handlers=[rich.logging.RichHandler(level=log_level)],
    )

    rich.traceback.install(suppress=["pep517", "pip"])



import hashlib
import shutil
import sys
import tarfile
import time
import zipfile
from pathlib import Path

import pytest

from scikit_build_core.build import build_sdist, build_wheel

DIR = Path(__file__).parent.resolve()
HELLO_PEP518 = DIR / "packages/simple_pyproject_ext"
ENTRYPOINTS = """\
[one.two]
three = four

[console_scripts]
something = other

[gui_scripts]
guithing = a.b:c
"""
METADATA = """\
Metadata-Version: 2.1
Name: cmake-example
Version: 0.0.1
Requires-Python: >=3.7
Provides-Extra: test
Requires-Dist: pytest>=6.0; extra == "test"
"""

mark_hashes_different = pytest.mark.xfail(
    sys.platform.startswith("win32") or sys.platform.startswith("cygwin"),
    reason="hashes differ on Windows",
    strict=False,
)


def test_pep517_sdist(tmp_path, monkeypatch):

    dist = tmp_path.resolve() / "dist"
    monkeypatch.chdir(HELLO_PEP518)
    if Path("dist").is_dir():
        shutil.rmtree("dist")
    out = build_sdist(str(dist))

    (sdist,) = dist.iterdir()
    assert "cmake-example-0.0.1.tar.gz" == sdist.name
    assert sdist == dist / out

    with tarfile.open(sdist) as f:
        file_names = set(f.getnames())
        assert file_names == {
            f"cmake-example-0.0.1/{x}"
            for x in (
                "CMakeLists.txt",
                "pyproject.toml",
                "src/main.cpp",
                "PKG-INFO",
            )
        }
        pkg_info = f.extractfile("cmake-example-0.0.1/PKG-INFO")
        assert pkg_info
        pkg_info_contents = pkg_info.read().decode()
        assert pkg_info_contents == METADATA


@mark_hashes_different
def test_pep517_sdist_hash(tmp_path, monkeypatch):
    dist = tmp_path.resolve() / "dist"
    monkeypatch.chdir(HELLO_PEP518)
    if Path("dist").is_dir():
        shutil.rmtree("dist")
    out = build_sdist(str(dist))
    sdist = dist / out
    hash = hashlib.sha256(sdist.read_bytes()).hexdigest()
    if sys.version_info < (3, 9):
        assert (
            hash == "600ed996e51642027557759ee9eeb31b5cae1f443313f5f7d0a40d9cc9cbdd13"
        )
    else:
        assert (
            hash == "4f47a4e797db1cb8e15afb368360d5f2ac5ae4b6c7e38e0771f8eba65fab65e4"
        )


def test_pep517_sdist_time_hash(tmp_path, monkeypatch):
    dist = tmp_path.resolve() / "dist"
    monkeypatch.chdir(HELLO_PEP518)
    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_sdist(str(dist))
    sdist = dist / out
    hash1 = hashlib.sha256(sdist.read_bytes()).hexdigest()

    time.sleep(2)
    Path("src/main.cpp").touch()

    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_sdist(str(dist))
    sdist = dist / out

    hash2 = hashlib.sha256(sdist.read_bytes()).hexdigest()

    assert hash1 == hash2


def test_pep517_sdist_time_hash_nonreproducable(tmp_path, monkeypatch):
    dist = tmp_path.resolve() / "dist"
    monkeypatch.chdir(HELLO_PEP518)
    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_sdist(str(dist), {"sdist.reproducible": "false"})
    sdist = dist / out
    hash1 = hashlib.sha256(sdist.read_bytes()).hexdigest()

    time.sleep(2)

    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_sdist(str(dist))
    sdist = dist / out

    hash2 = hashlib.sha256(sdist.read_bytes()).hexdigest()

    assert hash1 != hash2


@mark_hashes_different
def test_pep517_sdist_time_hash_set_epoch(tmp_path, monkeypatch):
    dist = tmp_path.resolve() / "dist"
    monkeypatch.chdir(HELLO_PEP518)
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "12345")
    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_sdist(str(dist), {"sdist.reproducible": "true"})
    sdist = dist / out
    hash = hashlib.sha256(sdist.read_bytes()).hexdigest()
    if sys.version_info < (3, 9):
        assert (
            hash == "505cb72c11e9b8344e6d467aef94f3e96d66d1c618a0703e4fcdbb623f28c23c"
        )
    else:
        assert (
            hash == "68703d101d8185d86ec7496285ddf46e302166c60a10372682de82f70bba847a"
        )


@pytest.mark.compile
@pytest.mark.configure
def test_pep517_wheel(tmp_path, monkeypatch, virtualenv):
    dist = tmp_path / "dist"
    dist.mkdir()
    monkeypatch.chdir(HELLO_PEP518)
    if Path("dist").is_dir():
        shutil.rmtree("dist")
    out = build_wheel(str(dist))
    (wheel,) = dist.glob("cmake_example-0.0.1-*.whl")
    assert wheel == dist / out

    if sys.version_info >= (3, 8):
        with wheel.open("rb") as f:
            p = zipfile.Path(f)
            file_names = [p.name for p in p.iterdir()]
            metadata = p.joinpath("cmake_example-0.0.1.dist-info/METADATA").read_text()
            entry_points = p.joinpath(
                "cmake_example-0.0.1.dist-info/entrypoints.txt"
            ).read_text()

        assert len(file_names) == 2
        assert "cmake_example-0.0.1.dist-info" in file_names
        file_names.remove("cmake_example-0.0.1.dist-info")
        (so_file,) = file_names

        assert so_file.startswith("cmake_example")
        print("SOFILE:", so_file)

        print(entry_points == ENTRYPOINTS)
        assert 'Requires-Dist: pytest>=6.0; extra == "test"' in metadata
        assert "Metadata-Version: 2.1" in metadata
        assert "Name: cmake-example" in metadata
        assert "Version: 0.0.1" in metadata
        assert "Requires-Python: >=3.7" in metadata
        assert "Provides-Extra: test" in metadata

    virtualenv.install(wheel)

    version = virtualenv.execute(
        "import cmake_example; print(cmake_example.__version__)",
    )
    assert version.strip() == "0.0.1"

    add = virtualenv.execute(
        "import cmake_example; print(cmake_example.add(1, 2))",
    )
    assert add.strip() == "3"


@pytest.mark.skip(reason="Doesn't work yet")
@pytest.mark.compile
@pytest.mark.configure
def test_pep517_wheel_time_hash(tmp_path, monkeypatch):
    dist = tmp_path / "dist"
    dist.mkdir()
    monkeypatch.chdir(HELLO_PEP518)
    monkeypatch.setenv("SOURCE_DATE_EPOCH", "12345")
    if Path("dist").is_dir():
        shutil.rmtree("dist")
    out = build_wheel(str(dist))
    wheel = dist / out
    hash1 = hashlib.sha256(wheel.read_bytes()).hexdigest()

    time.sleep(2)
    Path("src/main.cpp").touch()

    if Path("dist").is_dir():
        shutil.rmtree("dist")

    out = build_wheel(str(dist))
    wheel = dist / out

    hash2 = hashlib.sha256(wheel.read_bytes()).hexdigest()

    assert hash1 == hash2



from __future__ import annotations

import sys

if sys.version_info < (3, 9):
    import importlib_resources as resources
else:
    from importlib import resources

__all__ = ["resources"]


def __dir__() -> list[str]:
    return __all__



import dataclasses
from typing import Dict, List, Optional

__all__ = [
    "ScikitBuildSettings",
    "NinjaSettings",
    "CMakeSettings",
    "LoggingSettings",
    "SDistSettings",
    "BackportSettings",
    "WheelSettings",
]


def __dir__() -> List[str]:
    return __all__


@dataclasses.dataclass
class NinjaSettings:
    #: The minimum version of Ninja to use. If Ninja is older than this, it will
    #: be upgraded via PyPI if possible. An empty string will disable this check.
    minimum_version: str = "1.5"

    #: If make is present, do not add ninja if missing.
    make_fallback: bool = True


@dataclasses.dataclass
class CMakeSettings:
    #: The minimum version of CMake to use. If CMake is older than this, it will
    #: be upgraded via PyPI if possible. An empty string will disable this check.
    minimum_version: str = "3.15"

    #: A list of args to pass to CMake when configuring the project.
    args: List[str] = dataclasses.field(default_factory=list)

    #: A table of defines to pass to CMake when configuring the project. Additive.
    define: Dict[str, str] = dataclasses.field(default_factory=dict)

    #: Verbose printout when building
    verbose: bool = False

    #: The build type to use when building the project.
    #: Valid options are: "Debug", "Release", "RelWithDebInfo", "MinSizeRel",
    #: "", etc.
    build_type: str = "Release"


@dataclasses.dataclass
class LoggingSettings:
    #: The logging level to display.
    level: str = "WARNING"


@dataclasses.dataclass
class SDistSettings:
    #: Files to include in the SDist even if they are skipped by default.
    include: List[str] = dataclasses.field(default_factory=list)

    #: Files to exclude from the SDist even if they are included by default.
    exclude: List[str] = dataclasses.field(default_factory=list)

    #: If set to True, try to build a reproducible distribution.
    #: SOURCE_DATE_EPOCH will be used for timestamps, or a fixed value if not set.
    reproducible: bool = True


@dataclasses.dataclass
class WheelSettings:
    #: A list of packages to auto-copy into the wheel. If this is None, it will
    #: default to the first of ``src/<package>`` or ``<package>`` if they exist.
    #: The prefix(s) will be stripped from the package name inside the wheel.
    packages: Optional[List[str]] = None

    #: The Python tags. The default (empty string) will use the default
    #: Python version. You can also set this to "cp37" to enable the CPython
    #: 3.7+ Stable ABI / Limited API (only on CPython and if the version is
    #: sufficient, otherwise this has no effect). Or you can set it to "py3"
    #: or "py2.py3" to ignore Python ABI compatibility. For the stable ABI,
    #: the CMake variable SKBUILD_SOABI will be set to abi3 on Unix-like systems
    #: (empty on Windows). FindPython doesn't have a way to target python3.dll
    #: instead of python3N.dll, so this is harder to use on Windows.
    #: The ABI tag is inferred from this tag.
    py_api: str = ""

    #: Fill out extra tags that are not required. This adds "x86_64" and "arm64"
    #: to the list of platforms when "universal2" is used, which helps older
    #: Pip's (before 21.0.1) find the correct wheel.
    expand_macos_universal_tags: bool = False

    #: The install directory for the wheel. This is relative to the platlib root.
    #: EXPERIMENTAL: An absolute path will be one level higher than the platlib
    #: root, giving access to "/platlib", "/data", "/headers", and "/scripts".
    install_dir: str = ""


@dataclasses.dataclass
class BackportSettings:
    #: If CMake is less than this value, backport a copy of FindPython. Set
    #: to 0 disable this, or the empty string.
    find_python: str = "3.24"


@dataclasses.dataclass
class ScikitBuildSettings:
    cmake: CMakeSettings = dataclasses.field(default_factory=CMakeSettings)
    ninja: NinjaSettings = dataclasses.field(default_factory=NinjaSettings)
    logging: LoggingSettings = dataclasses.field(default_factory=LoggingSettings)
    sdist: SDistSettings = dataclasses.field(default_factory=SDistSettings)
    wheel: WheelSettings = dataclasses.field(default_factory=WheelSettings)
    backport: BackportSettings = dataclasses.field(default_factory=BackportSettings)

    #: Strictly check all config options. If False, warnings will be
    #: printed for unknown options. If True, an error will be raised.
    strict_config: bool = True

    #: Enable early previews of features not finalized yet.
    experimental: bool = False

    #: If set, this will provide a method for backward compatibility.
    minimum_version: Optional[str] = None



from __future__ import annotations

from pathlib import Path

import nox

DIR = Path(__file__).parent.resolve()

nox.options.sessions = ["lint", "pylint", "tests"]


@nox.session(reuse_venv=True)
def lint(session: nox.Session) -> None:
    """
    Run the linter.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session(reuse_venv=True)
def pylint(session: nox.Session) -> None:
    """
    Run PyLint.
    """
    # This needs to be installed into the package environment, and is slower
    # than a pre-commit check
    session.install("-e.[dev,test]", "pylint")
    session.run("pylint", "src", *session.posargs)


@nox.session(reuse_venv=True)
def tests(session: nox.Session) -> None:
    """
    Run the unit and regular tests.
    """
    env = {"PIP_DISABLE_PIP_VERSION_CHECK": "1"}
    session.install("-e.[test]")
    session.run("pytest", *session.posargs, env=env)


@nox.session(reuse_venv=True)
def coverage(session: nox.Session) -> None:
    """
    Run coverage and report.
    """

    session.install("-e.[test,cov]")
    session.run("pytest", "--cov=scikit_build_core", *session.posargs)


@nox.session
def docs(session: nox.Session) -> None:
    """
    Build the docs. Pass "--serve" to serve.
    """

    session.install(".[docs]")
    session.chdir("docs")
    session.run("sphinx-build", "-M", "html", ".", "_build")

    if session.posargs:
        if "--serve" in session.posargs:
            print("Launching docs at http://localhost:8000/ - use Ctrl-C to quit")
            session.run("python", "-m", "http.server", "8000", "-d", "_build/html")
        else:
            session.warn("Unsupported argument to docs")


@nox.session
def build(session: nox.Session) -> None:
    """
    Build an SDist and wheel.
    """

    session.install("build")
    session.run("python", "-m", "build", **session.posargs)



from __future__ import annotations

import dataclasses
import os
import subprocess
from collections.abc import Iterable

from ._logging import logger

__all__ = ["Run"]


def __dir__() -> list[str]:
    return __all__


@dataclasses.dataclass
class Run:
    env: dict[str, str] | None = None
    cwd: os.PathLike[str] | None = None

    def live(self, *args: str | os.PathLike[str]) -> None:
        """
        Runs code and prints the results live.
        """
        self._run(args, capture=False)

    def capture(
        self, *args: str | os.PathLike[str]
    ) -> subprocess.CompletedProcess[str]:
        """
        Runs a command and captures the result.
        """
        return self._run(args, capture=True)

    def _run(
        self,
        args: Iterable[str | os.PathLike[str]],
        capture: bool,
    ) -> subprocess.CompletedProcess[str]:

        options = [
            os.fspath(arg) if isinstance(arg, os.PathLike) else arg for arg in args
        ]

        logger.debug("RUN: {}", " ".join(options))

        return subprocess.run(
            options,
            text=True,
            check=True,
            capture_output=capture,
            env=self.env,
            cwd=self.cwd,
        )



from __future__ import annotations

import contextlib
import copy
import gzip
import io
import os
import tarfile
from pathlib import Path

from pyproject_metadata import StandardMetadata

from .._compat import tomllib
from ..settings.skbuild_read_settings import SettingsReader
from ._file_processor import each_unignored_file
from ._init import setup_logging

__all__: list[str] = ["build_sdist"]


def __dir__() -> list[str]:
    return __all__


def get_reproducible_epoch() -> int:
    """
    Return an integer representing the integer number of seconds since the Unix epoch.

    If the `SOURCE_DATE_EPOCH` environment variable is set, use that value. Otherwise,
    always return `1667997441`.
    """
    return int(os.environ.get("SOURCE_DATE_EPOCH", "1667997441"))


def normalize_file_permissions(st_mode: int) -> int:
    """
    Normalize the permission bits in the st_mode field from stat to 644/755
    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.

    Taken from https://github.com/pypa/flit/blob/6a2a8c6462e49f584941c667b70a6f48a7b3f9ab/flit_core/flit_core/common.py#L257
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode


def normalize_tar_info(tar_info: tarfile.TarInfo) -> tarfile.TarInfo:
    """
    Normalize the TarInfo associated with a file to improve reproducibility.

    Inspired by Hatch
    https://github.com/pypa/hatch/blob/573192f88022bb781c698dae2c0b84ef3fb9a7ad/backend/src/hatchling/builders/sdist.py#L51
    """
    tar_info = copy.copy(tar_info)
    tar_info.uname = ""
    tar_info.gname = ""
    tar_info.uid = 0
    tar_info.gid = 0
    tar_info.mode = normalize_file_permissions(tar_info.mode)
    tar_info.mtime = get_reproducible_epoch()
    return tar_info


def build_sdist(
    sdist_directory: str,
    # pylint: disable-next=unused-argument
    config_settings: dict[str, list[str] | str] | None = None,
) -> str:
    settings_reader = SettingsReader(Path("pyproject.toml"), config_settings or {})
    settings = settings_reader.settings
    setup_logging(settings.logging.level)

    settings_reader.validate_may_exit()

    sdist_dir = Path(sdist_directory)

    with Path("pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)

    reproducible = settings.sdist.reproducible
    timestamp = get_reproducible_epoch() if reproducible else None

    metadata = StandardMetadata.from_pyproject(pyproject)
    pkg_info = bytes(metadata.as_rfc822())

    srcdirname = f"{metadata.name}-{metadata.version}"
    filename = f"{srcdirname}.tar.gz"

    sdist_dir.mkdir(parents=True, exist_ok=True)
    with contextlib.ExitStack() as stack:
        gzip_container = stack.enter_context(
            gzip.GzipFile(sdist_dir / filename, mode="wb", mtime=timestamp)
        )
        tar = stack.enter_context(
            tarfile.TarFile(fileobj=gzip_container, mode="w", format=tarfile.PAX_FORMAT)
        )
        for filepath in each_unignored_file(
            Path("."), include=settings.sdist.include, exclude=settings.sdist.exclude
        ):
            tar.add(
                filepath,
                arcname=srcdirname / filepath,
                filter=normalize_tar_info if reproducible else lambda x: x,
            )

        tarinfo = tarfile.TarInfo(name=f"{srcdirname}/PKG-INFO")
        tarinfo.size = len(pkg_info)
        if reproducible:
            tarinfo = normalize_tar_info(tarinfo)
        with io.BytesIO(pkg_info) as fileobj:
            tar.addfile(tarinfo, fileobj)

    return filename
